<?php

namespace Iwaydev\Yii2Sender;

use DomainException;
use GuzzleHttp\Client;
use GuzzleHttp\RequestOptions;
use JsonException;
use Psr\Http\Message\ResponseInterface;
use Throwable;
use yii\base\Component;
use yii\caching\CacheInterface;
use Yii;

/**
 * Сендер для отправки запросов
 *
 * Class SenderComponent
 *
 * @package Iwaydev\Yii2Sender
 */
class SenderComponent extends Component
{
    /** @var string URI запроса */
    public string $url;
    /** @var string Версия используемого API */
    public ?string $version = null;
    /** @var string|null Название кэш компоента */
    public ?string $cacheComponent = null;
    /** @var int Время жизни кэша в секундах */
    public int $cacheLifetime = 3600;
    /** @var string|null Категория логов по-умолчанию */
    public ?string $defaultLogCategory;
    /** @var string|null XUUID */
    public ?string $xUuid = null;

    /** @var array Заголовки */
    protected array $headers = [];

    /** @var bool Возвращать ли ошибки в ответе */
    private bool $httpErrors = false;
    /** @var Client Клиент */
    private Client $client;
    /** @var CacheInterface|null Инстанс кэшера */
    private ?CacheInterface $cache = null;
    /** @var bool Флаг кэширования */
    private bool $caching = false;
    /** @var string|null Категрия логов */
    private ?string $logCategory = null;

    /**
     * SenderComponent constructor.
     *
     * @param array $config Конфиги
     */
    public function __construct($config = [])
    {
        parent::__construct($config);
        $this->client = new Client([
            'base_uri' => $this->url,
            'http_errors' => false
        ]);
        $this->version = trim($this->version, '/');

        $this->setXUuid($this->xUuid);
        $this->setCache();
    }

    /**
     * Возвращает текущий http-клиент
     *
     * @return Client
     */
    public function getClient(): Client
    {
        return $this->client;
    }

    /**
     * Установить кэш
     *
     * @return $this
     */
    public function setCache()
    {
        if (empty($this->cacheComponent)) {
            return $this;
        }
        $cache = Yii::$app->get($this->cacheComponent);
        if (!($cache instanceof CacheInterface)) {
            return $this;
        }
        $this->cache = $cache;

        return $this;
    }

    /**
     * Возвращать ошибки из ответа
     *
     * @return $this
     */
    public function withErrors(): self
    {
        $this->httpErrors = true;

        return $this;
    }

    /**
     * Отправить GET-запрос
     *
     * @param string $uri URI запроса
     * @param array $query GET-параметры
     *
     * @return array
     * @throws JsonException
     */
    public function get(string $uri, array $query = []): array
    {
        return $this->send('GET', $uri, ['query' => $query]);
    }

    /**
     * Отправить POST-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    public function post(string $uri, array $body = []): array
    {
        return $this->send('POST', $uri, ['form_params' => $body]);
    }

    /**
     * Отправить PATCH-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    public function patch(string $uri, array $body = []): array
    {
        return $this->send('PATCH', $uri, ['form_params' => $body]);
    }

    /**
     * Отправить PUT-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    public function put(string $uri, array $body = []): array
    {
        return $this->send('PUT', $uri, ['form_params' => $body]);
    }

    /**
     * Отправить DELETE-запрос
     *
     * @param string $uri URI запроса
     * @param array $body Данные запроса
     *
     * @return array
     * @throws JsonException
     */
    public function delete(string $uri, array $body = []): array
    {
        return $this->send('DELETE', $uri, ['form_params' => $body]);
    }

    /**
     * Кэшировать следующий запрос
     *
     * @return $this
     */
    public function caching(): self
    {
        $this->caching = true;

        return $this;
    }

    /**
     * Отправка запроса
     *
     * @param string $method Тип запроса GET, POST и пр.
     * @param string $uri URI запроса
     * @param array $options Параметры запроса
     *
     * @return array
     * @throws JsonException
     */
    private function send(string $method, string $uri, array $options = []): array
    {
        try {
            $options = array_merge($options, ['headers' => $this->headers]);
            $uri = $this->prepareUri($uri);
            $fullUrl = $this->client->getConfig('base_uri') . $uri;

            $this->logInfo('Request. Method: ' . $method);
            $this->logInfo('Request. Url: ' . $fullUrl);
            $this->logInfo('Request. Options:', $options);

            $key = (string)$fullUrl . json_encode($options);

            if (!empty($this->cache) && $this->caching === true) {
                $result = $this->cache->get($key);
                if (!empty($result)) {
                    $this->logInfo('Response. Is caching!');
                    $this->logInfo('Response. Content:', $result);

                    return $result;
                }
            }
            if ($this->httpErrors) {
                $options[RequestOptions::HTTP_ERRORS] = true;
                $this->httpErrors = false;
            }

            $response = $this->client->request($method, $uri, $options);
            $result = $this->prepareResponse($response);

            if (!empty($this->cache) && $this->caching === true) {
                $this->cache->set($key, $result, $this->cacheLifetime);
                $this->caching = false;
            }

            $this->logInfo('Response. Content:', $result);

            $this->setLogCategory($this->defaultLogCategory);

            return $result;
        } catch (Throwable $throwable) {
            $this->logException($throwable);

            throw $throwable;
        }
    }

    /**
     * Подготовка uri
     *
     * @param string $uri URI запроса
     *
     * @return string
     */
    protected function prepareUri(string $uri): string
    {
        return DIRECTORY_SEPARATOR . $this->prepareVersion() . trim($uri, '/');
    }

    /**
     * Возвращает версию апи, подготовленную для подстановки в uri
     *
     * @return string
     */
    private function prepareVersion(): string
    {
        if (empty($this->version)) {
            return '';
        }

        return $this->version . DIRECTORY_SEPARATOR;
    }

    /**
     * Подготовка ответа
     *
     * @param ResponseInterface $response Инстанс ответа
     *
     * @return array
     * @throws JsonException
     */
    protected function prepareResponse(ResponseInterface $response): array
    {
        return json_decode(
            $response->getBody(),
            true,
            512,
            JSON_THROW_ON_ERROR
        );
    }

    /**
     * Устанавливает категорию логов
     *
     * @param string|null $logCategory Категория логов
     *
     * @return $this
     */
    public function setLogCategory(?string $logCategory): self
    {
        $this->logCategory = $logCategory;

        if (empty($this->defaultLogCategory)) {
            $this->defaultLogCategory = $logCategory;
        }

        return $this;
    }

    /**
     * Сделать запись уровня "info" в лог, если категория логов установлена
     *
     * @param string $message Сообщение
     * @param array $context Дополнительные данные
     */
    private function logInfo(string $message, array $context = []): void
    {
        if (!empty($this->logCategory)) {
            Yii::info($this->xUuid . ' ' . $message . "/n" . json_encode($context), $this->logCategory);
        }
    }

    /**
     * Сделать запись уровня "error" в лог, если категория логов установлена
     *
     * @param string $message Сообщение
     * @param array $context Дополнительные данные
     */
    private function logError(string $message, array $context = []): void
    {
        if (!empty($this->logCategory)) {
            Yii::error($this->xUuid . ' ' . $message . "/n" . json_encode($context), $this->logCategory);
        }
    }

    /**
     * Логируем исключение
     *
     * @param Throwable $throwable Экзепляр исключения
     */
    public function logException(Throwable $throwable): void
    {
        $this->logError('');
        $this->logError('Исключение!');
        $this->logError('Сообщение: ' . $throwable->getMessage());
        $this->logError('Код: ' . $throwable->getCode());
        $this->logError('Файл: ' . $throwable->getFile() . '(' . $throwable->getLine() . ')');
        $this->logError('Трейс: ' . PHP_EOL . $throwable->getTraceAsString());
        $this->logError('');
    }

    /**
     * Устанавливает заголовок
     *
     * @param string $header Заголовок
     * @param string $value Значение
     *
     * @return $this
     */
    public function setHeader(string $header, string $value): self
    {
        $this->headers[$header] = $value;

        return $this;
    }

    /**
     * Возвращает заголовок, если он есть
     *
     * @param string $name
     *
     * @return string
     */
    public function getHeader(string $name): string
    {
        if (isset($this->headers[$name])) {
            return $this->headers[$name];
        }

        throw new DomainException('Header ' . $name . ' not found');
    }

    /**
     * Устанавливает либо генерирует заголовок X-UUID
     *
     * @param string|null $xUuid Значение которое будет установлено в качестве заголовка X-UUID
     *
     * @return $this
     */
    public function setXUuid(?string $xUuid = null): self
    {
        $this->xUuid = empty($xUuid) ? md5(microtime()) : $xUuid;

        $this->setHeader('X-UUID', $this->xUuid);

        return $this;
    }

    /**
     * Возвращает текущий заголовок X-UUID
     *
     * @return string
     */
    public function getXUuid(): string
    {
        return $this->getHeader('X-UUID');
    }
}
